@@ -1,7 +1,7 @@
 """
 Core functions used by unumpy and some of its submodules.
 
-(c) 2010-2013 by Eric O. LEBIGOT (EOL).
+(c) 2010-2016 by Eric O. LEBIGOT (EOL).
 """
 
 # The functions found in this module cannot be defined in unumpy or
@@ -12,17 +12,20 @@ Core functions used by unumpy and some of its submodules.
 from __future__ import division
 
 # Standard modules:
+from builtins import next
+from builtins import zip
+from builtins import range
 import sys
+import inspect
 
 # 3rd-party modules:
 import numpy
 from numpy.core import numeric
 
 # Local modules:
-import uncertainties
-from uncertainties import umath, deprecation
-
-from uncertainties import __author__
+import uncertainties.umath_core as umath_core
+import uncertainties.core as uncert_core
+from uncertainties.core import deprecation
 
 __all__ = [
     # Factory functions:
@@ -43,7 +46,7 @@ __all__ = [
 # might well directly build arrays of numbers with uncertainties
 # without going through the factory functions found in this module
 # (uarray() and umatrix()).  Thus,
-# numpy.array([uncertainties.ufloat((1, 0.1))]) would not
+# numpy.array([uncert_core.ufloat((1, 0.1))]) would not
 # have a nominal_values() method.  Adding such a method to, say,
 # unumpy.matrix, would break the symmetry between NumPy arrays and
 # matrices (no nominal_values() method), and objects defined in this
@@ -54,15 +57,15 @@ __all__ = [
 # nominal_values (i.e. the documentation of its class):
 
 to_nominal_values = numpy.vectorize(
-    uncertainties.nominal_value,
+    uncert_core.nominal_value,
     otypes=[float],  # Because vectorize() has side effects (dtype setting)
-    doc=("Applies uncertainties.nominal_value to the elements of"
-         " a NumPy (or unumpy) array (this includes matrices)."))    
+    doc=("Return the nominal value of the numbers with uncertainties contained"
+         " in a NumPy (or unumpy) array (this includes matrices)."))
 
 to_std_devs = numpy.vectorize(
-    uncertainties.std_dev,
-    otypes=[float],  # Because vectorize() has side effects (dtype setting)    
-    doc=("Returns the standard deviation of the numbers with uncertainties"
+    uncert_core.std_dev,
+    otypes=[float],  # Because vectorize() has side effects (dtype setting)
+    doc=("Return the standard deviation of the numbers with uncertainties"
          " contained in a NumPy array, or zero for other objects."))
 
 def unumpy_to_numpy_matrix(arr):
@@ -71,15 +74,19 @@ def unumpy_to_numpy_matrix(arr):
     Otherwise, it is returned unchanged.
     """
 
-    return arr.view(numpy.matrix) if isinstance(arr, matrix) else arr        
+    if isinstance(arr, matrix):
+        return arr.view(numpy.matrix)
+    else:
+        return arr
 
 def nominal_values(arr):
     """
-    Returns the nominal values of the numbers in NumPy array arr.
+    Return the nominal values of the numbers in NumPy array arr.
 
-    Elements that are not uncertainties.AffineScalarFunc are passed
-    through untouched (because a numpy.array can contain numbers with
-    uncertainties and pure floats simultaneously).
+    Elements that are not numbers with uncertainties (derived from a
+    class from this module) are passed through untouched (because a
+    numpy.array can contain numbers with uncertainties and pure floats
+    simultaneously).
 
     If arr is of type unumpy.matrix, the returned array is a
     numpy.matrix, because the resulting matrix does not contain
@@ -90,11 +97,12 @@ def nominal_values(arr):
 
 def std_devs(arr):
     """
-    Returns the standard deviations of the numbers in NumPy array arr.
+    Return the standard deviations of the numbers in NumPy array arr.
 
-    Elements that are not uncertainties.AffineScalarFunc are given a
-    zero uncertainty ((because a numpy.array can contain numbers with
-    uncertainties and pure floats simultaneously)..
+    Elements that are not numbers with uncertainties (derived from a
+    class from this module) are passed through untouched (because a
+    numpy.array can contain numbers with uncertainties and pure floats
+    simultaneously).
 
     If arr is of type unumpy.matrix, the returned array is a
     numpy.matrix, because the resulting matrix does not contain
@@ -107,11 +115,11 @@ def std_devs(arr):
 
 def derivative(u, var):
     """
-    Returns the derivative of u along var, if u is an
-    uncertainties.AffineScalarFunc instance, and if var is one of the
+    Return the derivative of u along var, if u is an
+    uncert_core.AffineScalarFunc instance, and if var is one of the
     variables on which it depends.  Otherwise, return 0.
     """
-    if isinstance(u, uncertainties.AffineScalarFunc):
+    if isinstance(u, uncert_core.AffineScalarFunc):
         try:
             return u.derivatives[var]
         except KeyError:
@@ -120,31 +128,37 @@ def derivative(u, var):
         return 0.
 
 def wrap_array_func(func):
+    # !!! This function is not used in the code, except in the tests.
+    #
+    # !!! The implementation seems superficially similar to
+    # uncertainties.core.wrap(): is there code/logic duplication
+    # (which should be removed)?
     """
-    Returns a version of the function func() that works even when
+    Return a version of the function func() that works even when
     func() is given a NumPy array that contains numbers with
-    uncertainties.
+    uncertainties, as first argument.
 
-    This wrapper is similar to uncertainties.wrap(), except that it
-    handles an array argument instead of float arguments.
+    This wrapper is similar to uncertainties.core.wrap(), except that
+    it handles an array argument instead of float arguments, and that
+    the result can be an array.
 
     However, the returned function is more restricted: the array
     argument cannot be given as a keyword argument with the name in
     the original function (it is not a drop-in replacement).
-    
-    func -- function whose first argument takes a single NumPy array,
+
+    func -- function whose first argument is a single NumPy array,
     and which returns a NumPy array.
     """
 
-    @uncertainties.set_doc("""\
+    @uncert_core.set_doc("""\
     Version of %s(...) that works even when its first argument is a NumPy
     array that contains numbers with uncertainties.
-    
+
     Warning: elements of the first argument array that are not
-    AffineScalarFunc objects must not depend on uncertainties.Variable
+    AffineScalarFunc objects must not depend on uncert_core.Variable
     objects in any way.  Otherwise, the dependence of the result in
-    uncertainties.Variable objects will be incorrect.
-    
+    uncert_core.Variable objects will be incorrect.
+
     Original documentation:
     %s""" % (func.__name__, func.__doc__))
     def wrapped_func(arr, *args, **kwargs):
@@ -159,8 +173,13 @@ def wrap_array_func(func):
         variables = set()
         for element in arr.flat:
             # floats, etc. might be present
-            if isinstance(element, uncertainties.AffineScalarFunc):
-                variables |= set(element.derivatives.iterkeys())
+            if isinstance(element, uncert_core.AffineScalarFunc):
+                # !!!! The following forces an evaluation of the
+                # derivatives!? Isn't this very slow, when
+                # working with a large number of arrays?
+                #
+                # !! set() is only needed for Python 2 compatibility:
+                variables |= set(element.derivatives.keys())
 
         # If the matrix has no variables, then the function value can be
         # directly returned:
@@ -196,7 +215,7 @@ def wrap_array_func(func):
             # The standard deviation might be numerically too small
             # for the evaluation of the derivative, though: we set the
             # minimum variable shift.
-            
+
             shift_var = max(var._std_dev/1e5, 1e-8*abs(var._nominal_value))
             # An exceptional case is that of var being exactly zero.
             # In this case, an arbitrary shift is used for the
@@ -219,13 +238,26 @@ def wrap_array_func(func):
             # derivatives, for each element:
             for (derivative_dict, derivative_value) in (
                 zip(derivatives.flat, numerical_deriv.flat)):
-                
+
                 if derivative_value:
                     derivative_dict[var] = derivative_value
 
         # numbers with uncertainties are built from the result:
-        return numpy.vectorize(uncertainties.AffineScalarFunc)(
-            func_nominal_value, derivatives)
+        return numpy.vectorize(uncert_core.AffineScalarFunc)(
+            func_nominal_value,
+            numpy.vectorize(uncert_core.LinearCombination)(derivatives))
+
+    wrapped_func = uncert_core.set_doc("""\
+    Version of %s(...) that works even when its first argument is a NumPy
+    array that contains numbers with uncertainties.
+
+    Warning: elements of the first argument array that are not
+    AffineScalarFunc objects must not depend on uncert_core.Variable
+    objects in any way.  Otherwise, the dependence of the result in
+    uncert_core.Variable objects will be incorrect.
+
+    Original documentation:
+    %s""" % (func.__name__, func.__doc__))(wrapped_func)
 
     # It is easier to work with wrapped_func, which represents a
     # wrapped version of 'func', when it bears the same name as
@@ -239,7 +271,7 @@ def wrap_array_func(func):
 
 def uarray(nominal_values, std_devs=None):
     """
-    Returns a NumPy array of numbers with uncertainties
+    Return a NumPy array of numbers with uncertainties
     initialized with the given nominal values and standard
     deviations.
 
@@ -255,29 +287,29 @@ def uarray(nominal_values, std_devs=None):
     if std_devs is None:  # Obsolete, single tuple argument call
         deprecation('uarray() should now be called with two arguments.')
         (nominal_values, std_devs) = nominal_values
-        
+
     return (numpy.vectorize(
-        # ! Looking up uncertainties.Variable beforehand through
-        # '_Variable = uncertainties.Variable' does not result in a
+        # ! Looking up uncert_core.Variable beforehand through
+        # '_Variable = uncert_core.Variable' does not result in a
         # significant speed up:
-        lambda v, s: uncertainties.Variable(v, s), otypes=[object])
+        lambda v, s: uncert_core.Variable(v, s), otypes=[object])
         (nominal_values, std_devs))
 
 ###############################################################################
 
 def array_derivative(array_like, var):
     """
-    Returns the derivative of the given array with respect to the
+    Return the derivative of the given array with respect to the
     given variable.
 
-    The returned derivative is a Numpy ndarray of the same shape as
+    The returned derivative is a NumPy ndarray of the same shape as
     array_like, that contains floats.
 
     array_like -- array-like object (list, etc.)  that contains
     scalars or numbers with uncertainties.
 
     var -- Variable object.
-    """    
+    """
     return numpy.vectorize(lambda u: derivative(u, var),
                            # The type is set because an
                            # integer derivative should not
@@ -286,30 +318,34 @@ def array_derivative(array_like, var):
                            otypes=[float])(array_like)
 
 def func_with_deriv_to_uncert_func(func_with_derivatives):
+    # This function is used for instance for the calculation of the
+    # inverse and pseudo-inverse of a matrix with uncertainties.
     """
-    Returns a function that can be applied to array-like objects that
-    contain numbers with uncertainties (lists, lists of lists, Numpy
+    Return a function that can be applied to array-like objects that
+    contain numbers with uncertainties (lists, lists of lists, NumPy
     arrays, etc.).
-    
-    func_with_derivatives -- defines a function that takes array-like
-    objects containing scalars and returns an array.  Both the value
-    and the derivatives of this function with respect to multiple
-    scalar parameters are calculated by func_with_derivatives().
-    
+
+    func_with_derivatives -- defines a function that takes an
+    array-like object containing scalars and returns an array.  Both
+    the value and the derivatives of this function with respect to
+    multiple scalar parameters are calculated by this
+    func_with_derivatives() argument.
+
     func_with_derivatives(arr, input_type, derivatives, *args,
-    **kwargs) returns an iterator.  The first element is the value of
-    the function at point 'arr' (with the correct type).  The
-    following elements are arrays that represent the derivative of the
-    function for each derivative array from the iterator
-    'derivatives'.
+    **kwargs) must return an iterator.  The first element returned by
+    this iterator is the value of the function at the n-dimensional
+    array-like 'arr' (with the correct type).  The following elements
+    are arrays that represent the derivative of the function for each
+    derivative array from the iterator 'derivatives'.
 
-      func_with_derivatives takes the following arguments:
+    func_with_derivatives() takes the following arguments:
 
-      arr -- Numpy ndarray of scalars where the function must be
+      arr -- NumPy ndarray of scalars where the function must be
       evaluated.
 
-      input_type -- type of the input array-like object.  This type is
-      used for determining the type that the function should return.
+      input_type -- data type of the input array-like object.  This
+      type is used for determining the type that the function should
+      return.
 
       derivatives -- iterator that returns the derivatives of the
       argument of the function with respect to multiple scalar
@@ -321,50 +357,64 @@ def func_with_deriv_to_uncert_func(func_with_derivatives):
 
     Examples of func_with_derivatives: inv_with_derivatives().
     """
-    
+
     def wrapped_func(array_like, *args, **kwargs):
         """
-        array_like -- array-like object that contains numbers with
-        uncertainties (list, Numpy ndarray or matrix, etc.).
+        array_like -- n-dimensional array-like object that contains
+        numbers with uncertainties (list, NumPy ndarray or matrix,
+        etc.).
 
         args -- additional arguments that are passed directly to
         func_with_derivatives.
         """
 
-        # So that .flat works even if array_like is a list.  Later
-        # useful for faster code:
-        array_version = numpy.asarray(array_like)
+        # The calculation below is not lazy, contrary to the linear
+        # error propagation done in AffineScalarFunc. Making it lazy
+        # in the same way would be quite a specific task: basically
+        # this would amount to generalizing scalar coefficients in
+        # core.LinearCombination to more general matrix
+        # multiplications, and to replace Variable differentials by
+        # full matrices of coefficients. This does not look very
+        # efficient, as matrices are quite big, and since caching the
+        # result of a few matrix functions that are not typically
+        # stringed one after the other (unlike a big sum of numbers)
+        # should not be needed.
+
+        # So that .flat works even if array_like is a list:
+        array_version = numpy.asanyarray(array_like)
 
         # Variables on which the array depends are collected:
         variables = set()
         for element in array_version.flat:
             # floats, etc. might be present
-            if isinstance(element, uncertainties.AffineScalarFunc):
-                variables |= set(element.derivatives.iterkeys())
+            if isinstance(element, uncert_core.AffineScalarFunc):
+                # !!! set() is only needed for Python 2 compatibility:
+                variables |= set(element.derivatives.keys())
 
         array_nominal = nominal_values(array_version)
-        # Function value, and derivatives at array_nominal (the
+        # Function value, then derivatives at array_nominal (the
         # derivatives are with respect to the variables contained in
         # array_like):
-        func_and_derivs = func_with_derivatives(
+        func_then_derivs = func_with_derivatives(
             array_nominal,
             type(array_like),
             (array_derivative(array_version, var) for var in variables),
             *args, **kwargs)
 
-        func_nominal_value = func_and_derivs.next()
+        func_nominal_value = next(func_then_derivs)
 
         if not variables:
             return func_nominal_value
-        
+
         # The result is built progressively, with the contribution of
         # each variable added in turn:
 
         # Calculation of the derivatives of the result with respect to
         # the variables.
-        derivatives = numpy.array(
-            [{} for _ in xrange(func_nominal_value.size)], dtype=object)
-        derivatives.resize(func_nominal_value.shape)
+        derivatives = (
+            numpy.array(
+                [{} for _ in range(func_nominal_value.size)], dtype=object)
+            .reshape(func_nominal_value.shape))
 
         # Memory-efficient approach.  A memory-hungry approach would
         # be to calculate the matrix derivatives will respect to all
@@ -373,7 +423,8 @@ def func_with_deriv_to_uncert_func(func_with_derivatives):
         # progressively build the matrix of derivatives, by
         # progressively adding the derivatives with respect to
         # successive variables.
-        for (var, deriv_wrt_var) in zip(variables, func_and_derivs):
+        for (var, deriv_wrt_var) in zip(variables,
+                                                   func_then_derivs):
 
             # Update of the list of variables and associated
             # derivatives, for each element:
@@ -382,18 +433,19 @@ def func_with_deriv_to_uncert_func(func_with_derivatives):
                 if derivative_value:
                     derivative_dict[var] = derivative_value
 
-        # An array of numbers with uncertainties are built from the
+        # An array of numbers with uncertainties is built from the
         # result:
-        result = numpy.vectorize(uncertainties.AffineScalarFunc)(
-            func_nominal_value, derivatives)
+        result = numpy.vectorize(uncert_core.AffineScalarFunc)(
+            func_nominal_value,
+            numpy.vectorize(uncert_core.LinearCombination)(derivatives))
 
-        # Numpy matrices that contain numbers with uncertainties are
+        # NumPy matrices that contain numbers with uncertainties are
         # better as unumpy matrices:
         if isinstance(result, numpy.matrix):
             result = result.view(matrix)
-                    
+
         return result
-    
+
     return wrapped_func
 
 ########## Matrix inverse
@@ -429,7 +481,7 @@ inv.__doc__ = """\
 
     The result is a unumpy.matrix if numpy.linalg.pinv would return a
     matrix for the array of nominal values.
-    
+
     Analytical formulas are used.
 
     Original documentation:
@@ -486,15 +538,27 @@ def pinv_with_derivatives(arr, input_type, derivatives, rcond):
         yield term1+term2+term3
 
 # Default rcond argument for the generalization of numpy.linalg.pinv:
-try:
-    # Python 2.6+:
-    pinv_default = numpy.linalg.pinv.__defaults__[0]
-except AttributeError:
-    pinv_default = 1e-15
+#
+# Most common modern case first:
+try: 
+    pinv_default = (
+        inspect.signature(numpy.linalg.pinv).parameters["rcond"].default)
+except AttributeError:  # No inspect.signature() before Python 3.3
+    try:
+        # In numpy 1.17+, pinv is wrapped using a decorator which unfortunately
+        # results in the metadata (argument defaults) being lost. However, we
+        # can still get at the original function using the __wrapped__
+        # attribute (which is what inspect.signature() does).
+        pinv_default = numpy.linalg.pinv.__wrapped__.__defaults__[0]
+    except AttributeError:  # Function not wrapped in NumPy < 1.17
+        pinv_default = numpy.linalg.pinv.__defaults__[0]  # Python 1, 2.6+:
 
 pinv_with_uncert = func_with_deriv_to_uncert_func(pinv_with_derivatives)
 
-@uncertainties.set_doc("""
+def pinv(array_like, rcond=pinv_default):
+    return pinv_with_uncert(array_like, rcond)
+
+pinv = uncert_core.set_doc("""
     Version of numpy.linalg.pinv that works with array-like objects
     that contain numbers with uncertainties.
 
@@ -505,12 +569,32 @@ pinv_with_uncert = func_with_deriv_to_uncert_func(pinv_with_derivatives)
 
     Original documentation:
     %s
-    """ % numpy.linalg.pinv.__doc__)
-def pinv(array_like, rcond=pinv_default):
-    return pinv_with_uncert(array_like, rcond)
+    """ % numpy.linalg.pinv.__doc__)(pinv)
 
 ########## Matrix class
 
+class CallableStdDevs(numpy.matrix):
+    '''
+    Class for standard deviation results, which used to be
+    callable. Provided for compatibility with old code. Issues an
+    obsolescence warning upon call.
+
+    New objects must be created by passing an existing
+    '''
+
+    def __new__(cls, matrix):
+        # The following prevents a copy of the original matrix, which
+        # could be expensive, and is unnecessary (the CallableStdDevs
+        # is just a wrapping around the original matrix, which can be
+        # modified):
+        matrix.__class__ = cls
+        return matrix
+
+    def __call__ (self):
+        deprecation('the std_devs attribute should not be called'
+                    ' anymore: use .std_devs instead of .std_devs().')
+        return self
+
 class matrix(numpy.matrix):
     # The name of this class is the same as NumPy's, which is why it
     # does not follow PEP 8.
@@ -520,7 +604,7 @@ class matrix(numpy.matrix):
     """
 
     def __rmul__(self, other):
-        # ! NumPy's matrix __rmul__ uses an apparently a restrictive
+        # ! NumPy's matrix __rmul__ uses an apparently restrictive
         # dot() function that cannot handle the multiplication of a
         # scalar and of a matrix containing objects (when the
         # arguments are given in this order).  We go around this
@@ -530,31 +614,33 @@ class matrix(numpy.matrix):
         else:
             return numeric.dot(other, self)  # The order is important
 
-    # The NumPy doc for getI is empty:
-    # @uncertainties.set_doc(numpy.matrix.getI.__doc__)
     def getI(self):
-        "Matrix inverse or pseudo-inverse"
-        
-        # numpy.matrix.getI is OK too, but the rest of the code assumes that
-        # numpy.matrix.I is a property object anyway:
-
+        """Matrix inverse or pseudo-inverse."""
         m, n = self.shape
         return (inv if m == n else pinv)(self)
 
-    # ! In Python >= 2.6, this could be simplified as:
-    # I = numpy.matrix.I.getter(__matrix_inverse)
-    I = property(getI, numpy.matrix.I.fset, numpy.matrix.I.fdel,
-                 numpy.matrix.I.__doc__)
+    I = numpy.matrix.I.getter(getI)
 
+
+    # !!! The following function is not in the official documentation
+    # of the module. Maybe this is because arrays with uncertainties
+    # do not have any equivalent in this module, and they should be
+    # the first ones to have such methods?
     @property
     def nominal_values(self):
         """
         Nominal value of all the elements of the matrix.
         """
         return nominal_values(self)
-    
-    std_devs = std_devs
-    
+
+    # !!! The following function is not in the official documentation
+    # of the module. Maybe this is because arrays with uncertainties
+    # do not have any equivalent in this module, and they should be
+    # the first ones to have such methods?
+    @property
+    def std_devs(self):
+        return CallableStdDevs(std_devs(self))
+
 def umatrix(nominal_values, std_devs=None):
     """
     Constructs a matrix that contains numbers with uncertainties.
@@ -569,14 +655,14 @@ def umatrix(nominal_values, std_devs=None):
     if std_devs is None:  # Obsolete, single tuple argument call
         deprecation('umatrix() should now be called with two arguments.')
         (nominal_values, std_devs) = nominal_values
-            
+
     return uarray(nominal_values, std_devs).view(matrix)
 
 ###############################################################################
 
 def define_vectorized_funcs():
     """
-    Defines vectorized versions of functions from uncertainties.umath.
+    Defines vectorized versions of functions from uncertainties.umath_core.
 
     Some functions have their name translated, so as to follow NumPy's
     convention (example: math.acos -> numpy.arccos).
@@ -585,39 +671,58 @@ def define_vectorized_funcs():
     this_module = sys.modules[__name__]
     # NumPy does not always use the same function names as the math
     # module:
-    func_name_translations = dict(
+    func_name_translations = dict([
         (f_name, 'arc'+f_name[1:])
-        for f_name in ['acos', 'acosh', 'asin', 'atan', 'atan2', 'atanh'])
+        for f_name in ['acos', 'acosh', 'asin', 'atan', 'atan2', 'atanh']])
+
+    new_func_names = [
+        func_name_translations.get(function_name, function_name)
+        # The functions from umath_core.non_std_wrapped_funcs
+        # (available from umath) are normally not in
+        # NumPy, so they are not included here:
+        for function_name in umath_core.many_scalars_to_scalar_funcs]
 
-    new_func_names = [func_name_translations.get(function_name, function_name)
-                      # The functions from umath.non_std_wrapped_funcs
-                      # (available from umath) are normally not in
-                      # numpy, so they are not included here:
-                      for function_name in umath.many_scalars_to_scalar_funcs]
-        
     for (function_name, unumpy_name) in zip(
-        umath.many_scalars_to_scalar_funcs, new_func_names):
+        umath_core.many_scalars_to_scalar_funcs, new_func_names):
 
         # ! The newly defined functions (uncertainties.unumpy.cos, etc.)
         # do not behave exactly like their NumPy equivalent (numpy.cos,
         # etc.): cos(0) gives an array() and not a
         # numpy.float... (equality tests succeed, though).
-        func = getattr(umath, function_name)
+        func = getattr(umath_core, function_name)
+
+        # Data type of the result of the unumpy function:
+        otypes = (
+            # It is much more convenient to preserve the type of
+            # functions that return a number without
+            # uncertainty. Thus, for example, unumpy.isnan() can
+            # return an array with a boolean data type (instead of
+            # object), which allows the result to be used with NumPy's
+            # boolean indexing.
+            {} if function_name in umath_core.locally_cst_funcs
+            # If by any chance a function returns, in a particular
+            # case, an integer instead of a number with uncertainty,
+            # side-effects in vectorize() would fix the resulting
+            # dtype to integer, which is not what is wanted (as
+            # vectorize(), at least in NumPy around 2010 maybe,
+            # decided about the output data type by looking at the
+            # type of first element only).
+            else {'otypes': [object]})
+
         setattr(
             this_module, unumpy_name,
+            #!!!! For umath_core.locally_cst_funcs, would it make sense
+            # to optimize this by using instead the equivalent (? see
+            # above) vectorized NumPy function on the nominal values?
             numpy.vectorize(func,
-                            # If by any chance a function returns,
-                            # in a particular case, an integer,
-                            # side-effects in vectorize() would
-                            # fix the resulting dtype to integer,
-                            # which is not what is wanted:
-                            otypes=[object],
                             doc="""\
 Vectorized version of umath.%s.
 
 Original documentation:
-%s""" % (function_name, func.__doc__)))
+%s""" % (function_name, func.__doc__),
+                            **otypes))
+
 
         __all__.append(unumpy_name)
-    
+
 define_vectorized_funcs()
